"""provide gitlab api used to read repositories"""
import os
import re

from urllib.parse import urljoin


DEFAULT_GIT_LAB_URL_FMT = re.compile(r"(?P<host>.+com|org|net)/(?P<project>.+)")

API_FORMAT = "%s/api/v%s/projects/%s/"  # hostname/api/:apiversion/projects/:id/

GIT_LAB_URL_FMT = re.compile(os.getenv("GIT_LAB_URL_FMT", DEFAULT_GIT_LAB_URL_FMT))


class RepoApi(object):
    """Used to generate the gitlab api used to access the config files.config
    It will parse the repo url to get according api with right host, id and
    branch.
    Users can also set their own regular expression format if the host doesn't end
    with .com/org/net
    The default api version is 4
    """

    def __init__(self, repo_url, v_3=False):
        m = GIT_LAB_URL_FMT.match(repo_url)
        if not m:
            raise ValueError(
                "The hostname must end with '.com|org|net' \
or set GIT_LAB_URL_FMT by yourself"
            )
        self.host = m.group("host")
        self.repo_id = m.group("project").replace("/", "%2F")
        self.v_3 = v_3

    @property
    def api_url(self):
        if self.v_3:
            return API_FORMAT % (self.host, "3", self.repo_id)
        return API_FORMAT % (self.host, "4", self.repo_id)

    def get_branch_list_api(self):
        return urljoin(self.api_url, "repository/branches")

    def get_file_list_api(self, page, branch=None):
        """Return the api used to get the list of files in the repo or files
        in the sub-module.
        """
        branch = branch.replace("/", "%2F") if branch else "master"
        return urljoin(self.api_url, "repository/tree?ref=%s&page=%s" % (branch, page))

    def get_file_content_api(self, file_path, branch=None):
        """Get the api used to retrieve the raw content.
        The input file_path should be from get_files_list_api or the full path to the file,
        for example:
            The url is 'https://some-host.com/ubi-config-data/tree/master/data/ubi7/config.yaml'
            Then the file path should data/ubi7/config.yaml
        """
        branch = branch.replace("/", "%2F") if branch else "master"
        encoded_file_path = file_path.replace("/", "%2F")
        return urljoin(
            self.api_url, "repository/files/%s/raw?ref=%s" % (encoded_file_path, branch)
        )



class Flag:
    def __init__(self, name, value):
        self.name = name
        self.value = _str_to_bool(value)


class Flags(object):
    """group of modules"""

    def __init__(self, flags):
        """
        Args:
            include(list): a list of :class:`Flag` instances
        """
        self._flags = flags

    def __getattr__(self, name):
        # allow getting flags as object attrs
        for item in self._flags:
            if item.name == name:
                return item

    def as_dict(self):
        return {item.name: item.value for item in self._flags}

    @classmethod
    def load_from_dict(cls, data):
        """Create instances of :class:`Flags` from a dictionary

        Args:
            data(dict): dictionary with data of following format

        .. code-block:: json

            {
                "flag_1": "value",
                "flag_2": "value"
            }
        """
        return cls([Flag(name, value) for name, value in data.items()])


def _str_to_bool(value):
    out = value

    if isinstance(value, str):
        if value.lower() in ("true", "yes"):
            out = True
        elif value.lower() in ("false", "no"):
            out = False

    return out



import os
import json

from jsonschema import validate

DEFAULT_SCHEMA = os.path.join(os.path.dirname(__file__), "config_schema.json")
""" Default yaml schema used for validation of UbiConfig configuration files """


def validate_config(data, schema=None):
    """Validate the data according to the schema
    If no schema is provided, :data:`DEFAULT_SCHEMA` is used"""
    if schema is None:
        with open(DEFAULT_SCHEMA) as f:
            schema = json.load(f)
    validate(data, schema)



"""This module abstract the 'modules' content type"""


class Module(object):
    """Define a single module"""

    def __init__(self, name, stream, profiles=None):
        self.name = name
        self.stream = str(stream)
        self.profiles = profiles

    def __repr__(self):
        return "<Module: %s>" % self.name


class Modules(object):
    """group of modules"""

    def __init__(self, include):
        """
        Args:
            include(list): a list of :class:`Module` instances
        """
        self.whitelist = include

    def __getitem__(self, index):
        return self.whitelist[index]

    @classmethod
    def load_from_dict(cls, data):
        """Create instances of :class:`Modules` from a dictionary

        Args:
            data(dict): dictionary with data of following format

        .. code-block:: json

            {"include": [
                    {
                        "name": "<module_name>",
                        "stream": "<module_stream>",
                        "profiles": "<module_profiles>"
                    }
                ]
            }
        """
        include = []

        for item in data.get("include", []):
            module = Module(**item)
            include.append(module)

        return cls(include)



class Loader(object):
    """Load UBI configuration.

    Don't construct instances of this class directly;
    use the :func:`~ubiconfig.get_loader` function.
    """

    # This is a base class only. See implementations alongside this one
    # under the loaders module.

    def load(self, file_name, version=None):
        """
        Load a single configuration file and return a :class:`UbiConfig` object.

        The given file_name should be a relative path to a YAML file
        in the loader's config source (e.g. relative path to file within a
        git repo or local directory).

        version is an option argument to specify the version of config file to load,
        if it's None, then use default, e.g. ubi7, ubi8
        """
        raise NotImplementedError()

    def load_all(self):
        """Get the list of config files from repo and call load on every file.
        Return a list of :class:`UbiConfig` objects.
        """
        raise NotImplementedError()



import logging
import os
import re

import yaml
from jsonschema.exceptions import ValidationError

from ubiconfig.utils.config_validation import validate_config
from ubiconfig.config_types import UbiConfig


LOG = logging.getLogger("ubiconfig")


class LocalLoader(object):
    """Load configuration from a local directory tree."""

    def __init__(self, path):
        """
        Args:
            path (str): a local path to config files
        """

        self._path = path
        self._isroot = False
        # when load_all is called, the path returned includes self._path, no
        # need to join again.
        self._ver_files_map = self._get_local_files_map()
        # a {version: [file]} map

    def load(self, file_name, version=None):
        """Load a config file from local.

        Args:
            file_name (str): path to the config file.

            version(str):
                The version usage here is a little different from the remote loader,
                it's only used to denote the version of config file, but in remote
                loader, it's also used to find the right config.

                If version is None, we should get it from its path.
        """
        if not self._isroot:
            file_path = os.path.join(self._path, file_name)
        else:
            file_path = file_name

        if version is None:
            # get version from path, such as configs/ubi7.1/config.yaml, get
            # ubi7.1.
            version = os.path.basename(os.path.dirname(os.path.abspath(file_path)))

        if not re.search(r"ubi[0-9]\.[0-9]{1,2}$|ubi[0-9]$", version):
            raise ValueError(
                "Expect directories named in format ubi[0-9].([0-9]{1,2})$' or ubi[0-9]$, but got %s"
                % version
            )

        LOG.info("Loading configuration file locally: %s", file_path)

        with open(file_path, "r") as f:
            config_dict = yaml.load(f, Loader=yaml.BaseLoader)
        # validate input data
        validate_config(config_dict)

        return UbiConfig.load_from_dict(config_dict, file_name, version[3:])

    def load_all(self):
        """Load all config file from a local directory and all its subdirectories"""

        ubi_configs = []
        self._isroot = True

        for version, files in self._ver_files_map.items():
            for f in files:
                LOG.debug("Now loading %s", f)
                try:
                    ubi_configs.append(self.load(f, version))
                except yaml.YAMLError:
                    LOG.error(
                        "%s FAILED loading because of Syntax error, Skip for now", f
                    )
                    continue
                except ValidationError as e:
                    LOG.error("%s FAILED schema validation:\n%s\nSkip for now", f, e)
                    continue

        self._isroot = False
        # restore _isroot so the loader can be used again.

        return ubi_configs

    def _get_local_files_map(self):
        """Get the config file list from local."""
        LOG.info("Getting the local config file list")
        ver_files_map = {}
        for root, _, files in os.walk(self._path):
            conf_files = [
                os.path.join(root, f) for f in files if f.endswith((".yaml", ".yml"))
            ]
            if conf_files:
                # if there's yaml files, then it must under some version directory
                version = os.path.basename(os.path.abspath(root))
                ver_files_map.setdefault(version, []).extend(conf_files)
                # the result map is as {'version': ['file1', 'file2', ..]}

        return ver_files_map



import logging
import re
import yaml
import requests

from jsonschema.exceptions import ValidationError

from ubiconfig.utils.api.gitlab import RepoApi
from ubiconfig.utils.config_validation import validate_config
from ubiconfig.config_types import UbiConfig

from .base import Loader

LOG = logging.getLogger("ubiconfig")

BRANCH_RE = re.compile(r"^(?P<prefix>[\w-]{1,25})(?P<default_version>[\d]{1,2})")


class GitlabLoader(Loader):
    """Load configuration from a remote repo on gitlab."""

    def __init__(self, url):
        """
        :param url: gitlab repo url in form of `https://<host>/<repo>`
        """
        self._url = url
        self._session = requests.Session()
        self._repo_api = RepoApi(self._url.rstrip("/"))
        self._branches = self._get_branches()
        self._files_branch_map = self._pre_load()

    def load(self, file_name, version=None):
        """Load file from remote repository.
        :param file_name: filename that is on remote repository in any branch
        :param version: name of remote branch
        """
        if version is None:
            raise ValueError(
                "Provide valid name of remote branch, provided %s" % version
            )

        if file_name not in self._files_branch_map:
            raise ValueError(
                "Couldn't find file %s from remote repo %s" % (file_name, self._url)
            )

        match = re.match(BRANCH_RE, version)
        if not match:
            raise ValueError("Invalid version (branch name) %s" % version)

        prefix = match.group("prefix")
        default_version = match.group("default_version")

        default_branch = f"{prefix}{default_version}"

        sha1 = None
        loaded_version = None

        for branch_name in (version, default_branch):
            sha1 = self._branches.get(branch_name)
            if sha1:
                loaded_version = branch_name.lstrip(prefix)
                break

        if sha1 is None:
            raise ValueError(
                "Couldn't find version %s and default branch %s from %s for %s"
                % (version, default_branch, self._url, file_name)
            )

        LOG.info("Loading config file %s from branch %s", file_name, version)
        config_file_url = self._repo_api.get_file_content_api(file_name, sha1)
        response = self._session.get(config_file_url)
        response.raise_for_status()

        config_dict = yaml.load(response.content, Loader=yaml.BaseLoader)
        # validate input data
        validate_config(config_dict)

        return UbiConfig.load_from_dict(config_dict, file_name, loaded_version)

    def load_all(self):
        ubi_configs = []
        for f in self._files_branch_map:
            for branch_sha1 in self._files_branch_map[f]:
                LOG.debug("Now loading %s from branch %s", f, branch_sha1[0])
                try:
                    ubi_configs.append(self.load(f, branch_sha1[0]))
                except yaml.YAMLError:
                    LOG.error(
                        "%s FAILED loading because of Syntax error, skipping for now", f
                    )
                    continue
                except ValidationError as e:
                    LOG.error("%s FAILED schema validation:\n%s\nSkip for now", f, e)
                    continue

        return ubi_configs

    def _pre_load(self):
        """Iterate all branches to get a mapping of {file_path: (branch, sha1)...}"""
        files_branch_map = {}

        LOG.debug("Loading config files from all branches")

        for branch, sha1 in self._branches.items():
            page = 1
            while True:
                file_list_api = self._repo_api.get_file_list_api(branch=sha1, page=page)
                response = self._session.get(file_list_api)
                response.raise_for_status()
                file_list = [
                    f["path"]
                    for f in response.json()
                    if f["name"].endswith((".yaml", ".yml"))
                ]
                for f in file_list:
                    files_branch_map.setdefault(f, []).append((branch, sha1))
                    # now the map is {filename: [(branch1, sha1), (branch2, sha1),...]}
                    # same file name could map to multiple config files.
                if page >= int(response.headers.get("X-Total-Pages", 1)):
                    break
                page += 1

        return files_branch_map

    def _get_branches(self):
        """Get a {branch: sha1} mapping for all branches of a given repo"""
        branch_sha1 = {}

        LOG.info("Getting branches of the repo")
        branches_list_api = self._repo_api.get_branch_list_api()
        json_response = self._session.get(branches_list_api).json()
        if not json_response:
            raise RuntimeError("Please check %s is in right format" % self._url)
        for b in json_response:
            branch_sha1[b["name"]] = b["commit"]["id"]

        return branch_sha1



from .modules import Modules
from .packages import Packages
from .content_sets import ContentSetsMapping
from .flags import Flags


class UbiConfig(object):
    """Wrap all UBI related configurations
    Examples to access different configurations:

    Modules:
      config.modules[0].whitelist[0].name
    Packages:
      config.packages.whitelist[0].name
      config.packages.blacklist[0].name
    ContentSets:
      config.content_sets.rpm.input
      config.content_sets.debuginfo.output"""

    def __init__(self, cs, pkgs, mds, file_name, version, flags):
        """
        :param cs: :class:`~ubiconfig.config_types.content_sets.ContentSetsMapping`
        :param pkgs: :class:`~ubiconfig.config_types.packages.Packages`
        :param mds: :class:`~ubiconfig.config_types.modules.Modules`
        :param filename: filename used as identifier
        :param flags: :class:`~ubiconfig.config_types.flags.Flags`
        """
        self.content_sets = cs
        self.packages = pkgs
        self.modules = mds
        self.file_name = file_name
        self.version = version
        self.flags = flags

    def __repr__(self):
        return self.file_name

    @classmethod
    def load_from_dict(cls, data, file_name, version=None):
        """Create new instance of UbiConfig and load it from provided dictonary with
        following format:

        .. code-block:: json

            {
                "modules":  {},
                "packages": {
                    "include": ["package-name-.*"],
                    "exclude": ["package-name-.*"],
                    "arches": ["arch"]
                },
                "content_sets": {},
            }

        See also :meth:`ubiconfig.config_types.content_sets.ContentSetsMapping.load_from_dict`
        :meth:`ubiconfig.config_types.modules.Modules.load_from_dict`
        """
        m_data = Modules.load_from_dict(data.get("modules", {}))
        pkgs = data.get("packages", {})
        pkgs_data = Packages(
            pkgs.get("include", []), pkgs.get("exclude", []), data.get("arches", [])
        )
        cs_map = ContentSetsMapping.load_from_dict(data["content_sets"])

        flags = Flags.load_from_dict(data.get("flags", {}))
        # use the simplified file name
        file_name = file_name.split("/")[-1]

        return cls(
            cs=cs_map,
            pkgs=pkgs_data,
            mds=m_data,
            file_name=file_name,
            version=version,
            flags=flags,
        )



"""This module abstracts the 'content_sets' mappings"""


class ContentSetMapping(object):
    def __init__(self, input_content, output_content):
        self.input = input_content
        self.output = output_content


class Rpm(ContentSetMapping):
    """Input-output rpm content sets mapping"""

    @property
    def type(self):
        return "rpm"


class Srpm(ContentSetMapping):
    """Input-output srpm content sets mapping"""

    @property
    def type(self):
        return "srpm"


class Debuginfo(ContentSetMapping):
    """Input-output debuginfo content sets mapping"""

    @property
    def type(self):
        return "debuginfo"


class ContentSetsMapping(object):
    def __init__(self, rpm, srpm, debuginfo):
        """
        Args:
            rpm(:class:`Rpm`): rpm content sets mapping
            srpm(:class:`Srpm`): srpm content sets mapping
            debuginfo(:class:`Debuginfo`): debuginfo content sets mapping
        """
        self.rpm = rpm
        self.srpm = srpm
        self.debuginfo = debuginfo

    @classmethod
    def load_from_dict(cls, data):
        """Create instances of :class:`ContentSetsMapping` from a dictionary

        Args:
            data(dict): dictionary with data of following format

        .. code-block:: json

            {
                "rpm": {"input": "<input_content_set>",
                        "output":"<output_content_set>"},
                "srpm": {"input": "<input_content_set>",
                         "output":"<output_content_set>"},
                "debuginfo": {"input": "<input_content_set>",
                              "output":"<output_content_set>"}
            }
        """
        rpm_data = data.get("rpm", {})
        rpm = Rpm(rpm_data.get("input", ""), rpm_data.get("output", ""))

        srpm_data = data.get("srpm", {})
        srpm = Srpm(srpm_data.get("input", ""), srpm_data.get("output", ""))

        debug_data = data.get("debuginfo", {})
        debuginfo = Debuginfo(debug_data.get("input", ""), debug_data.get("output", ""))

        return cls(rpm, srpm, debuginfo)

    def export_dict(self):
        """Return a dictionary such as {type: (input, outpt)}"""
        attributes = [self.rpm, self.srpm, self.debuginfo]
        return dict((m.type, (m.input, m.output)) for m in attributes)



from ubiconfig.ubi import get_loader
from ubiconfig._impl.loaders.base import Loader
from ubiconfig.config_types import UbiConfig
from ubiconfig.utils.config_validation import validate_config

__all__ = ["get_loader", "Loader", "UbiConfig", "validate_config"]



"""This module abstract the 'packages' content type"""


class Package(object):
    def __init__(self, package, arches):
        self.package = package
        if "." in package:
            name, arch = package.rsplit(".", 1)
            if arch in arches or arch == "*":
                self.name, self.arch = name, arch
        if not hasattr(self, "name"):
            self.name = package
            self.arch = None

    def __repr__(self):
        return "<Package: %s>" % self.package


class IncludePackage(Package):
    def __init__(self, package, arches):
        super(IncludePackage, self).__init__(package, arches)
        if "*" in self.name:
            raise ValueError("<name>*.<arch> is not supported in whitelist")


class ExcludePackage(Package):
    pass


class Packages(object):
    def __init__(self, include, exclude, arches):
        """
        Args:
            include(list): list of packages to whitelist
            exclude(list): list of packages to blacklist
            arches(list): list of arches of packages in whitelist and blacklist
        """
        self.whitelist, self.blacklist = [], []
        for package in include:
            self.whitelist.append(IncludePackage(package, arches))
        for package in exclude:
            self.blacklist.append(ExcludePackage(package, arches))
